#include "common_ch.h"
#include "../../common.h"

#include "common_func.c"  // Yes, it's a C file.

void worker(volatile struct vm_cpu2_dev_st *dev) {
  struct api_t *api = dev->api;

  // Wait for other parts of the system to finish initialization.
  while (!dev->power_off && !dev->cpu3_ready) {
    api->msleep(10);
  }

  dev->cpu1->cpu2_ready = true;

  // Wait for CPU1 to signal that the data is ready.
  while (!dev->power_off && !dev->cpu1->apu_finished) {
    api->msleep(10);
  }

  // Process the data a little.
  for (int i = 0; i < 32; i++) {
    int ram_bank = i / 8;
    int ram_addr = i % 8;
    u16 data = ram_load(api, dev->ram_port[ram_bank], ram_addr);
    data ^= 0x6666;
    ram_store(api, dev->ram_port[3], 0x40+i, data);
  }

  // Signal that the data is ready and exit.
  dev->data_ready = true;
  dev->power_off = true;
}

// This is a workaround since ld linker cannot do -pie (or more explicitly,
// cannot do lea [rel...]) on non-ELF targets. Ugh.
size_t get_start_address(void) {
  void *ret = __builtin_return_address(0);
  return (size_t)ret & ~0xfffUL;
}

size_t get_end_address(size_t start) {
  size_t p = start;
  for (;; p++) {
    unsigned char *pul = (unsigned char*)p;
    int counter = 0;
    while (*pul == 0x41) {
      pul++;
      counter++;
    }

    if (counter == 8) {
      return p;
    }
  }

  return 0;  // Will never happen.
}

void *get_worker_addr(void);
__asm(
  ".globl get_worker_addr\n"
  "get_worker_addr:\n"
  "  lea rax, worker[rip]\n"
  "  ret\n"
);

void spawn_thread(struct api_t *api, size_t start, size_t end) {
  // Allocate io_dev register set region.
  struct vm_cpu2_dev_st *vm_cpu2_dev = api->mmap(sizeof(struct vm_cpu2_dev_st), PROT_RW);
  vm_cpu2_dev->api = api;

  // Send region location to master.
  api->send_to_master(&vm_cpu2_dev, sizeof(void*));

  // Receive other initialization information.
  struct vm_ram_dev_st *vm_ram_dev[4];
  int counter = 0;
  counter += api->recv_from_master(&vm_cpu2_dev->cpu1, 8);
  for (int i = 0; i < 4; i++) {
    counter += api->recv_from_master(&vm_ram_dev[i], 8);
  }

  if (counter != 8 + 4 * 8) {
    api->puts("Master disconnected.");
    api->abort(1);
    return;
  }

  for (int i = 0; i < 4; i++) {
    vm_cpu2_dev->ram_port[i] = &vm_ram_dev[i]->port[2];
  }

  // Allocate new thread region and copy code there.
  size_t sz = end - start;
  void *new_home = api->mmap(sz, PROT_RWX);

  unsigned char *p_src = (unsigned char*)start;
  unsigned char *p_dst = (unsigned char*)new_home;
  for (size_t i = 0; i < sz; i++, p_src++, p_dst++) {
    *p_dst = *p_src;
  }

  // Spawn new thread.
  size_t diff = (size_t)new_home - start;
  size_t worker_addr = (size_t)get_worker_addr() + diff;

  pthread_t th;
  if (api->thread_create(&th, (void*(*)(void*))worker_addr, vm_cpu2_dev) != 0) {
    api->abort(1);
  }
}

// Standard stuff follows.
__attribute__((section(".start")))
unsigned char main(struct api_t *api, unsigned char chain_value) {
  size_t start = get_start_address();
  size_t end = get_end_address(start);
  spawn_thread(api, start, end);

  unsigned char challenge;
  if (api->recv_from_master(&challenge, 1) != 1) {
    api->puts("Master disconnected.");
    api->set_finished();
    return 0;
  }

  unsigned char new_chain_value = ((chain_value + challenge) * (chain_value - challenge)) & 0xff;
  api->ack(new_chain_value);
  return new_chain_value;
}

__attribute__((section(".end")))
unsigned long __end_marker = 0x4141414141414141UL;

