import random
import sys
from keystone import *
import os

__all__ = [ 'obfuscate', 'read_file' ]

# Could should expect argument in RDI (ptr).
CODE_TEMPLATE = """
  mov r8, rdi
  mov r9, rsi
  mov ecx, {payload_length}
  lea rsi, [rel payload]
  mov rdi, rsi

decode_loop:
  lodsb
  {obfuscation_code}
  stosb
  loop decode_loop

  mov rdi, r8
  mov rsi, r9
  push rax        ; Align the stack to 16-bytes
  call payload
  mov rdx, rax
  pop rax         ; Remove stack alignment

  mov ecx, {payload_length}
  lea rdi, [rel payload]
  xor al, al
  rep stosb

  mov rax, rdx
  ret

payload:
"""

PAYLOAD_PADDING = 4096  # TODO change to 4096
PAD_DATA = os.urandom(4096)

ks = Ks(KS_ARCH_X86, KS_MODE_64)
ks.syntax = KS_OPT_SYNTAX_NASM

def asm(code):
  return bytearray(ks.asm(code, 0)[0])

def op_add():
  imm = random.randint(0, 255)
  code = "add al, 0x%.2x" % imm

  def encf(b):
    return (b - imm) & 0xff

  return code, encf

def op_sub():
  imm = random.randint(0, 255)
  code = "sub al, 0x%.2x" % imm

  def encf(b):
    return (b + imm) & 0xff

  return code, encf

def op_xor():
  imm = random.randint(0, 255)
  code = "xor al, 0x%.2x" % imm

  def encf(b):
    return b ^ imm

  return code, encf

def op_ror():
  imm = random.randint(0, 7)
  code = "ror al, 0x%.2x" % imm

  def encf(b):
    return ((b << imm) | (b >> (8 - imm))) & 0xff

  return code, encf

def op_rol():
  imm = random.randint(0, 7)
  code = "rol al, 0x%.2x" % imm

  def encf(b):
    return ((b >> imm) | (b << (8 - imm))) & 0xff

  return code, encf

def op_nop1():
  return "nop", None

OPS = [
  op_add,
  op_sub,
  op_xor,
  op_ror,
  op_rol,
  op_nop1,
]

def get_obfuscation_schema(num):
  functions = []
  code = []

  for i in xrange(num):
    c, ef = random.choice(OPS)()
    if ef is not None:
      functions.append(ef)
    code.append(c)

  return code, functions[::-1]

def read_file(fname):
  with open(fname, "rb") as f:
    return bytearray(f.read())

def write_file(fname, d):
  with open(fname, "wb") as f:
    f.write(d)

def obfuscate(org_data, alg_length=20):
  if len(org_data) % PAYLOAD_PADDING != 0:
    pad_size = PAYLOAD_PADDING - len(org_data) % PAYLOAD_PADDING
    org_data += PAD_DATA[-pad_size:]
    if len(org_data) % PAYLOAD_PADDING != 0:
      sys.exit("you failed at padding math lol")

  code, functions = get_obfuscation_schema(alg_length)

  payload = bytearray(org_data)
  for i, b in enumerate(payload):
    for f in functions:
      b = f(b)
    payload[i] = b

  full_code = CODE_TEMPLATE.format(**{
      "payload_length": len(payload),
      "obfuscation_code": '\n'.join(code)
  })

  return asm(full_code) + payload

def main(argv):
  if len(argv) != 3:
    return "usage: obfuscator.py <input_file> <output_file>"

  in_file = argv[1]
  out_file = argv[2]

  in_data = read_file(in_file)
  out_data = obfuscate(in_data)

  write_file(out_file, out_data)
  return 0


if __name__ == "__main__":
  sys.exit(main(sys.argv))
